/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer. 
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution. 
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission. 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#ifndef SETTINGS_H
#define SETTINGS_H

#include <QtCore>
class Settings
{
public:
    Settings(const QString &file = QString())
        : fileName(file)
    {
        if (fileName.isEmpty())
            fileName = globalSettingsFile();
    }

    QVariant value(const QString &key, const QVariant &defaultValue = QVariant()) const
    {
        QSettings settings(fileName, QSettings::IniFormat);
        const QStringList configs = QCoreApplication::instance()->property("configs").toStringList();

        foreach(const QString &config, configs) {
            settings.beginGroup(config);
            const QVariant value = settings.value(key);
            if (!value.isNull()) {
                return value;
            }
            settings.endGroup();
        }
        QVariant var = settings.value(key, defaultValue);
        if (var == defaultValue) {
            QString fn = globalSettingsFile();
            if (fn != fileName) {
                qSwap(fn, fileName);
                var = value(key, defaultValue);
                qSwap(fn, fileName);
            }
        }
        return var;
    }

    static QString globalSettingsFile()
    {
        return QCoreApplication::applicationDirPath().append("/../settings.conf");
    }
private:
    mutable QString fileName;
};

#endif
