/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include "gleswidget.h"
#include <QtDebug>

GLWidget::GLWidget(QWidget *parent)
        :QGLWidget(parent), program(0), textureId(0), floorTextureID(0), currentSurfaceIndex(0), textureFbo(0), scene(0), rotation(180), eyeX(-5), eyeZ(9)
{
}

GLWidget::~GLWidget()
{
    delete textureFbo;
}

void GLWidget::drawCube(GLfloat size, QMatrix4x4& pmvMatrix)
{
    float l = size/2;
    QVector<QVector3D> rightFace;
    QVector<QVector3D> backFace;
    QVector<QVector3D> bottomFace;
    QVector<QVector3D> topFace;
    QVector<QVector3D> leftFace;
    QVector<QVector3D> frontFace;
    static GLfloat vertices[8][3] = { {-l, l, l},
                                {l, l, l},
                                {l, -l, l},
                                {-l, -l, l},
                                {-l, l, -l},
                                {l, l, -l},
                                {l, -l, -l},
                                {-l, -l, -l}};

    static GLfloat texCoords[] = {0, 1, 0, 0, 1, 0, 1, 1};

    program->bind();
    program->enableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->enableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->setAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION, texCoords, 2, 0);

    program->setUniformValue("pmvMatrix", pmvMatrix);
    program->setUniformValue("color", QColor(128, 128, 255, 255));
    program->setUniformValue("blendWithBackground", true);

    //right
    //glNormal3f(-1.0, 0.0, 0.0);
    rightFace.append(QVector3D(vertices[7][0], vertices[7][1], vertices[7][2]));
    rightFace.append(QVector3D(vertices[4][0], vertices[4][1], vertices[4][2]));
    rightFace.append(QVector3D(vertices[0][0], vertices[0][1], vertices[0][2]));
    rightFace.append(QVector3D(vertices[3][0], vertices[3][1], vertices[3][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION,rightFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    // back
    //glNormal3f(0.0, 0.0, 1.0);
    backFace.append(QVector3D(vertices[6][0], vertices[6][1], vertices[6][2]));
    backFace.append(QVector3D(vertices[5][0], vertices[5][1], vertices[5][2]));
    backFace.append(QVector3D(vertices[4][0], vertices[4][1], vertices[4][2]));
    backFace.append(QVector3D(vertices[7][0], vertices[7][1], vertices[7][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, backFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    //bottom
    //glNormal3f(0.0, 1.0, 0.0);
    bottomFace.append(QVector3D(vertices[5][0], vertices[5][1], vertices[5][2]));
    bottomFace.append(QVector3D(vertices[1][0], vertices[1][1], vertices[1][2]));
    bottomFace.append(QVector3D(vertices[0][0], vertices[0][1], vertices[0][2]));
    bottomFace.append(QVector3D(vertices[4][0], vertices[4][1], vertices[4][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, bottomFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    // top
    //glNormal3f(0.0, -1.0, 0.0);
    topFace.append(QVector3D(vertices[7][0], vertices[7][1], vertices[7][2]));
    topFace.append(QVector3D(vertices[3][0], vertices[3][1], vertices[3][2]));
    topFace.append(QVector3D(vertices[2][0], vertices[2][1], vertices[2][2]));
    topFace.append(QVector3D(vertices[6][0], vertices[6][1], vertices[6][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, topFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    //left
    //glNormal3f(1.0, 0.0, 0.0);
    leftFace.append(QVector3D(vertices[2][0], vertices[2][1], vertices[2][2]));
    leftFace.append(QVector3D(vertices[1][0], vertices[1][1], vertices[1][2]));
    leftFace.append(QVector3D(vertices[5][0], vertices[5][1], vertices[5][2]));
    leftFace.append(QVector3D(vertices[6][0], vertices[6][1], vertices[6][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, leftFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    // front
    //glNormal3f(0.0, 0.0, -1.0);
    frontFace.append(QVector3D(vertices[3][0], vertices[3][1], vertices[3][2]));
    frontFace.append(QVector3D(vertices[0][0], vertices[0][1], vertices[0][2]));
    frontFace.append(QVector3D(vertices[1][0], vertices[1][1], vertices[1][2]));
    frontFace.append(QVector3D(vertices[2][0], vertices[2][1], vertices[2][2]));
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, frontFace.constData());
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);

    program->disableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->disableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->release();
}

void GLWidget::initializeGL()
{
    glClearColor(0.5, 0.5, 0.8, 1.0);
#ifdef QT_OPENGL_ES
    glClearDepthf(1.0);
#else
    glClearDepth(1.0);
#endif
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);

    program = new QGLShaderProgram(this);
    if (!program->addShaderFromSourceFile(QGLShader::Vertex, ":/resources/surface_shader.vert")) {
        qDebug() << "Vertex shader not compiled " << program->log();
    }
    if (!program->addShaderFromSourceFile(QGLShader::Fragment, ":/resources/surface_shader.frag")) {
        qDebug() << "Fragment shader not compiled " << program->log();
    }
    program->bindAttributeLocation("vertex", VERTEX_ATTRIBUTE_LOCATION);
    program->bindAttributeLocation("texCoord", TEXCOORD_ATTRIBUTE_LOCATION);
    if (!program->link()) {
        qDebug() << "Couldn't link shaders " << program->log();
    }
    floorTextureID = bindTexture(::loadPixmap(":/resources/marble.png"), GL_TEXTURE_2D, GL_RGBA,
                                 QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
    prepareSurface();
}

void GLWidget::resizeGL(int w, int h)
{
    QMatrix4x4 projectionMatrix;
    float x = w/h;
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
}

void GLWidget::drawFloor(QMatrix4x4& pmvMatrix)
{
    const GLfloat vertices[] = { -8.0, 0.0, 8.0, 0.0, 1.0,
                                 -8.0, 0.0, -8.0, 0.0, 0.0,
                                 8.0, 0.0, -8.0, 1.0, 0.0,
                                 8.0, 0.0, 8.0, 1.0, 1.0 };
    program->bind();
    program->enableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, vertices, 3, 5 * sizeof(GLfloat));
    program->enableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->setAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION, &vertices[3], 2, 5 * sizeof(GLfloat));
    program->setUniformValue("pmvMatrix", pmvMatrix);
//    program->setUniformValue("blendWithAlpha", false);
    glBindTexture(GL_TEXTURE_2D, floorTextureID);
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
    program->disableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->disableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->release();
}
void GLWidget::drawSurface(QMatrix4x4 &pmvMatrix)
{
    program->bind();
    program->enableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->setAttributeArray(VERTEX_ATTRIBUTE_LOCATION, vertices.constData());
    program->enableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->setAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION, texCoords.constData());
    program->setUniformValue("pmvMatrix", pmvMatrix);
    program->setUniformValue("blendWithAlpha", true);
    program->setUniformValue("alpha", 0.8F);
    glBindTexture(GL_TEXTURE_2D, textureId);
    glDrawElements(GL_TRIANGLES, indices.count(), GL_UNSIGNED_INT, indices.constData());
    program->disableAttributeArray(VERTEX_ATTRIBUTE_LOCATION);
    program->disableAttributeArray(TEXCOORD_ATTRIBUTE_LOCATION);
    program->release();
}

void GLWidget::prepareSurface()
{
    QFile file(QString(":/resources/surface%1.obj").arg(currentSurfaceIndex));
    QString line;
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) 
        return;

    indices.clear();
    vertices.clear();
    texCoords.clear();

    QVector<QVector2D> tempTexCoords;
    QVector<GLint> texIndices;
    QTextStream stream(&file);

    while (!stream.atEnd()) {
        line = stream.readLine();
        QStringList l = line.split(" ");
        if (l[0] == "v") {
            vertices.append(QVector3D(l[1].toDouble(), l[2].toDouble(), l[3].toDouble()));

        } else if (l[0] == "vt") {
            tempTexCoords.append(QVector2D(l[1].toDouble(), l[2].toDouble()));
        } else if (l[0] == "f") {
            QStringList temp;
            for (int i = 1; i < l.count(); i++) {
                temp = l[i].split("/");
                indices.append(temp[0].toUInt() - 1);
                texIndices.append(temp[1].toUInt() - 1);
            }
        }
    }
    file.close();
    // copy the correct sequence of texCoords into the main texCoords array
    texCoords.resize(vertices.size());
    for (int i = 0; i < indices.count(); i++) {
        texCoords[indices[i]] = tempTexCoords[texIndices[i]];
    }
}

void GLWidget::paintGL()
{
    if (scene) {
       if (!textureFbo || textureFbo->size() != size()) {
            delete textureFbo;
            textureFbo = 0;
            textureFbo = new QGLFramebufferObject(size());
        }

        QPainter painter(textureFbo);
        Q_ASSERT(size() == scene->visibleSceneRect().size().toSize());
        scene->render(&painter, QRectF(0, 0, width(), height()), scene->visibleSceneRect());
        painter.end();
        resizeGL(width(), height());

        textureId = textureFbo->texture();
    }

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    QMatrix4x4 matrix;
    //matrix.ortho(-4.0, 4.0, -4.0, 4.0, -3.0, 16.0);
    matrix.perspective(currentSurfaceIndex < 2 ? 55 : 70, width()/height(), 2, 16);
    matrix.lookAt(QVector3D(eyeX, 1, eyeZ), QVector3D(0, 0, 0), QVector3D(0, 1, 0));

    glEnable(GL_STENCIL_TEST);
    glStencilFunc(GL_ALWAYS, 1, 1);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glDisable(GL_DEPTH_TEST);
    glColorMask(0,0,0,0);
    QMatrix4x4 matrix1(matrix);
    matrix1.rotate(30, 1, 0, 0);
    matrix1.translate(0, -2.0, 0);
    drawFloor(matrix1);

    glEnable(GL_DEPTH_TEST);
    glColorMask(1,1,1,1);
    glStencilFunc(GL_EQUAL, 1, 1);

    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

    QMatrix4x4 matrix2(matrix);
    matrix2.scale(1.0, -1.0, 1.0);
    matrix2.translate(0, currentSurfaceIndex < 2 ? 6 : 6.4, 0);
    matrix2.rotate(rotation, 0, 0, 1);
    drawSurface(matrix2);
    glDisable(GL_STENCIL_TEST);

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    drawFloor(matrix1);

    QMatrix4x4 matrix3(matrix);
    matrix3.translate(0, 1.0, 0);
    matrix3.rotate(rotation, 0, 0, 1);
    drawSurface(matrix3);

    glDisable(GL_BLEND);
}

