/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#ifndef GLWIDGET_H
#define GLWIDGET_H

#include <QtGui>
#include <QtOpenGL>
#include "scene.h"

class GLWidget : public QGLWidget
{
    Q_OBJECT

public:
    GLWidget(QWidget *parent = 0);
    ~GLWidget();

    void setScene(GraphicsScene *s) { scene = s; updateGL(); }
    void draw() { glDraw(); }
    void drawCube(GLfloat size, QMatrix4x4& pmvMatrix);

public slots:
    void moveCameraLeft() { eyeX--; updateGL(); }
    void moveCameraRight() { eyeX++; updateGL(); }
    void moveCameraForward() { eyeZ--; updateGL(); }
    void moveCameraBackward() { eyeZ++; updateGL(); }

    void nextSurface()
    {
        currentSurfaceIndex = (currentSurfaceIndex+1) % 4;
        rotation = currentSurfaceIndex == 2 ? -90 : 180;
        prepareSurface();
        updateGL();
    }

protected:
    void initializeGL();
    void paintGL();
    void resizeGL(int, int);

private:
    void drawSurface(QMatrix4x4&);
    void drawFloor(QMatrix4x4&);

    void prepareSurface();

    enum AttrLoc { 
        VERTEX_ATTRIBUTE_LOCATION = 0, 
        TEXCOORD_ATTRIBUTE_LOCATION 
    };
    QGLShaderProgram *program;
    GLint textureId;
    GLint floorTextureID;
    QVector<QVector3D> vertices;
    QVector<QVector2D> texCoords;
    QVector<GLint> indices;
    int currentSurfaceIndex;
    QGLFramebufferObject *textureFbo;
    GraphicsScene *scene;

    int rotation;
    int eyeX, eyeZ;
};

#endif // GLWIDGET_H
