/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include "glwidget.h"
#include <QtGui>
#include <QtDebug>

GLWidget::GLWidget(QWidget *parent)
    : QGLWidget(parent), renderFbo(0), textureFbo(0), scene(0), currentSurfaceIndex(0), floorTextureId(0)
{
}

GLWidget::~GLWidget()
{
    delete renderFbo;
    if (renderFbo != textureFbo)
        delete textureFbo;
    gluDeleteNurbsRenderer(theNurb);
    deleteTexture(floorTextureId);
    glDeleteLists(startList, 3);
}

void GLWidget::drawCube(int size)
{
    int l = size/2;

    static GLint vertices[8][3] = { {-l, l, l},
                                {l, l, l},
                                {l, -l, l},
                                {-l, -l, l},
                                {-l, l, -l},
                                {l, l, -l},
                                {l, -l, -l},
                                {-l, -l, -l}};

    static GLint texCoords[4][2] = {{0, 1}, {0, 0}, {1, 0}, {1, 1}};

    //right
    glNormal3f(-1.0, 0.0, 0.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[7]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[4]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[0]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[3]);
    glEnd();
    // back
    glNormal3f(0.0, 0.0, 1.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[6]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[5]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[4]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[7]);
    glEnd();

    //bottom
    glNormal3f(0.0, 1.0, 0.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[5]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[1]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[0]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[4]);
    glEnd();

    // top
    glNormal3f(0.0, -1.0, 0.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[7]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[3]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[2]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[6]);
    glEnd();

    //left
    glNormal3f(1.0, 0.0, 0.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[2]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[1]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[5]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[6]);
    glEnd();

    // front
    glNormal3f(0.0, 0.0, -1.0);
    glBegin(GL_QUADS);
        glTexCoord2iv(texCoords[0]); glVertex3iv(vertices[3]);
        glTexCoord2iv(texCoords[1]); glVertex3iv(vertices[0]);
        glTexCoord2iv(texCoords[2]); glVertex3iv(vertices[1]);
        glTexCoord2iv(texCoords[3]); glVertex3iv(vertices[2]);
    glEnd();
}

void GLWidget::initializeGL()
{
    glClearColor(0.5, 0.5, 0.8, 1.0);
    glClearDepth(1.0f);
    glClearStencil(0);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_TEXTURE_2D);
    glShadeModel(GL_SMOOTH);

    // Create display list for 3 different kinds of nurbs objects
    startList = glGenLists(3);

    theNurb = gluNewNurbsRenderer();

    //initialize NURBS surface 1
    GLfloat s1CtrlPoints[4][4][3] = { {{-6.0,-3.0, 0.0}, {-3.0, -3.0, 2.0}, {3.0, -3.0, 2.0 }, {6.0, -3.0, 0.0} },
                                    {{-5.0, -1.0, 0.0}, {-2.0, -1.0, 2.0}, {2.0, -1.0, 2.0 }, {7.0, -1.0, 0.0} },
                                    {{-5.0, 1.0, 0.0}, {-2.0, 1.0, 2.0}, {2.0, 1.0, 2.0 }, {7.0, 1.0, 0.0} },
                                    {{-6.0, 3.0, 0.0}, {-3.0, 3.0, 2.0}, {3.0, 3.0, 2.0 }, {6.0, 3.0, 0.0} }};
    GLfloat sKnots[8] = { 0.0, 0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 1.0 };
    GLfloat texKnots[4] = { 0.0, 0.0, 1.0, 1.0 };
    GLfloat texPts[2][2][2] = { {{0.0, 0.0}, {1.0, 0.0}}, {{0.0, 1.0}, {1.0, 1.0}} };
    glNewList(startList, GL_COMPILE);
        gluBeginSurface(theNurb);
        gluNurbsSurface(theNurb, 8, sKnots, 8, sKnots, 4*3, 3, &s1CtrlPoints[0][0][0], 4, 4, GL_MAP2_VERTEX_3);
        gluNurbsSurface(theNurb, 4, texKnots, 4, texKnots, 4, 2, &texPts[0][0][0], 2, 2, GL_MAP2_TEXTURE_COORD_2);
        gluEndSurface(theNurb);
    glEndList();

    //initialize NURBS surface 2
    GLfloat s2CtrlPoints[4][4][3] = { {{-6.0, -1.0, -4.0}, {-12.0, -3.0, 0.0}, {12.0, -3.0, 0.0 }, {6.0, -1.0, -4.0} },
                                    {{-6.0, 1.0, -4.0}, {-12.0, -1.0, 0.0}, {12.0, -1.0, 0.0 }, {6.0, 1.0, -4.0} },
                                    {{-6.0, 3.0, -4.0}, {-12.0, 1.0, 0.0}, {12.0, 1.0, 0.0 }, {6.0, 3.0, -4.0} },
                                    {{-6.0, 5.0, -4.0}, {-12.0, 3.0, 0.0}, {12.0, 3.0, 0.0 }, {6.0, 5.0, -4.0} }};
    glNewList(startList + 1, GL_COMPILE);
        gluBeginSurface(theNurb);
        gluNurbsSurface(theNurb, 8, sKnots, 8, sKnots, 4*3, 3, &s2CtrlPoints[0][0][0], 4, 4, GL_MAP2_VERTEX_3);
        gluNurbsSurface(theNurb, 4, texKnots, 4, texKnots, 4, 2, &texPts[0][0][0], 2, 2, GL_MAP2_TEXTURE_COORD_2);
        gluEndSurface(theNurb);
    glEndList();

    //initialize NURBS surface 3
    GLfloat s3CtrlPoints[4][4][3] = { {{-6.0, -3.0, 0.0}, {-4.0, -5.0, -4.0}, {4.0, -1.0, 4.0 }, {6.0, -3.0, 0.0} },
                                    {{-6.0, -1.0, 0.0}, {-4.0, -3.0, -4.0}, {4.0, 1.0, 4.0 }, {6.0, -1.0, 0.0} },
                                    {{-6.0, 1.0, 0.0}, {-4.0, -1.0, -4.0}, {4.0, 3.0, 4.0 }, {6.0, 1.0, 0.0} },
                                    {{-6.0, 3.0, 0.0}, {-4.0, 1.0, -4.0}, {4.0, 5.0, 4.0 }, {6.0, 3.0, 0.0} }};
    glNewList(startList + 2, GL_COMPILE);
        gluBeginSurface(theNurb);
        gluNurbsSurface(theNurb, 8, sKnots, 8, sKnots, 4*3, 3, &s3CtrlPoints[0][0][0], 4, 4, GL_MAP2_VERTEX_3);
        gluNurbsSurface(theNurb, 4, texKnots, 4, texKnots, 4, 2, &texPts[0][0][0], 2, 2, GL_MAP2_TEXTURE_COORD_2);
        gluEndSurface(theNurb);
    glEndList();

    floorTextureId = bindTexture(::loadPixmap(":/resources/marble.png"), GL_TEXTURE_2D, GL_RGBA,
                                 QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
}

void GLWidget::setScene(GraphicsScene *s)
{
    scene = s;
}

void GLWidget::resizeGL(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    windowAspectRatio = (GLdouble) w/ (GLdouble) h;
    glFrustum(-windowAspectRatio, +windowAspectRatio, -1.0, 1.0, 1.5, 16.0);
    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
}

// methods to save the restore the state of the current GL context before and after using QPainter
void GLWidget::saveGLState()
{
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
}

void GLWidget::restoreGLState()
{
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glPopAttrib();
}

void GLWidget::drawSurface()
{
    glBindTexture(GL_TEXTURE_2D, textureId);
    glCallList(startList + currentSurfaceIndex);
}

void GLWidget::drawFloor()
{
    glBindTexture(GL_TEXTURE_2D, floorTextureId);
    glBegin(GL_QUADS);
        glNormal3f(0.0, 1.0, 0.0);
        // Normal Pointing Up
        glTexCoord2f(0.0f, 1.0f); // Bottom Left Of Texture
        glVertex3f(-10.0, 0.0, 10.0); // Bottom Left Corner Of Floor

        glTexCoord2f(0.0f, 0.0f); // Top Left Of Texture
        glVertex3f(-10.0, 0.0,-10.0); // Top Left Corner Of Floor

        glTexCoord2f(1.0f, 0.0f); // Top Right Of Texture
        glVertex3f( 10.0, 0.0,-10.0); // Top Right Corner Of Floor

        glTexCoord2f(1.0f, 1.0f); // Bottom Right Of Texture
        glVertex3f( 10.0, 0.0, 10.0); // Bottom Right Corner Of Floor
    glEnd();
}

void GLWidget::paintGL()
{
    static const bool useFrameBufferBlit = false; // !qgetenv("GLFBO_NO_FB_BLIT").toInt();
    if (scene) {
       if (!renderFbo || renderFbo->size() != size()) {
            delete renderFbo;
            if (renderFbo != textureFbo)
                delete textureFbo;
            renderFbo = 0;
            textureFbo = 0;
            if (useFrameBufferBlit && QGLFramebufferObject::hasOpenGLFramebufferBlit()) {
                QGLFramebufferObjectFormat format;
                format.setSamples(4); // to enable antialiasing
                format.setAttachment(QGLFramebufferObject::CombinedDepthStencil);
                renderFbo = new QGLFramebufferObject(size(), format);
                textureFbo = new QGLFramebufferObject(size());
            } else {
                renderFbo = new QGLFramebufferObject(size());
                textureFbo = renderFbo;
            }
        }

        // save GL state before using QPainter to render
        saveGLState();
        QPainter painter(renderFbo);
        Q_ASSERT(size() == scene->visibleSceneRect().size().toSize());
        scene->render(&painter, QRectF(0, 0, width(), height()), scene->visibleSceneRect());
        painter.end();
        restoreGLState();
        // Painting to FBO does not restore the GL Viewport (QTBUG-6712)
        resizeGL(width(), height());

        if (renderFbo != textureFbo) {
            QRect rect(0, 0, renderFbo->width(), renderFbo->height());
            QGLFramebufferObject::blitFramebuffer(textureFbo, rect,
                                               renderFbo, rect);
        }

        textureId = textureFbo->texture();
    }

    glLoadIdentity();
    gluLookAt(-1.5, .5, 8.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

    double eqr[] = {0.0f,-2.5f, 0.0f, 0.0f};
    glColorMask(0, 0, 0, 0);
    glEnable(GL_STENCIL_TEST);
    glStencilFunc(GL_ALWAYS, 1, 1);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

    glDisable(GL_DEPTH_TEST);
    glPushMatrix();
        glTranslatef(0.0, -2.5, 0.0);
        drawFloor(); // draw floor to stencil buffer
    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
    glColorMask(1,1,1,1);

    glStencilFunc(GL_EQUAL, 1, 1);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    glEnable(GL_CLIP_PLANE0);

    // draw reflected object
    glClipPlane(GL_CLIP_PLANE0, eqr);
    glPushMatrix();
        glScalef(1.0f, -1.0f, 1.0f);
        glTranslatef(0.0f, 5.9, 0.0f);
        drawSurface(); // Draw reflection
    glPopMatrix();
    glDisable(GL_CLIP_PLANE0);
    glDisable(GL_STENCIL_TEST);

    // Draw floor
    glEnable(GL_BLEND);
    glColor4f(1.0f, 1.0f, 1.0f, 0.7f);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPushMatrix();
        glTranslatef(0.0, -2.5, 0.0);
        drawFloor();
    glPopMatrix();
    glDisable(GL_BLEND);

    // Draw real surface again
    glEnable(GL_BLEND);
    glColor4f(1.0f, 1.0f, 1.0f, 0.7f);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPushMatrix();
        glTranslatef(0.0f, 0.5, 0.0f);
        drawSurface();
    glPopMatrix();
    glDisable(GL_BLEND);

    glFlush();
}
