/****************************************************************************

This file is part of the Common-Demo project on http://qt.gitorious.org.

Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).*
All rights reserved.

Contact:  Nokia Corporation (qt-info@nokia.com)**

You may use this file under the terms of the BSD license as follows:

"Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of Nokia Corporation and its Subsidiary(-ies) nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."

****************************************************************************/

#include "scene.h"
#include "items.h"
#include "view.h"
#include <QtScript>
#include <QWebFrame>
#include <QMenu>
#include "settings.h"
#define  FRIENDLY_KEYS=1

enum {
    SpecialItemZ = 10000000,
    AlmostSpecialItemZ = SpecialItemZ-1,
    CurrentItemZ = 0xDEAD,
    DefaultItemZ = 0
};

class ScriptInterface : public QObject
{
    Q_OBJECT
    Q_PROPERTY(QString easingCurve READ easingCurve WRITE setEasingCurve)
    Q_PROPERTY(int duration READ duration WRITE setDuration)
    Q_PROPERTY(qreal overshoot READ overshoot WRITE setOvershoot)
    Q_PROPERTY(qreal amplitude READ amplitude WRITE setAmplitude)
    Q_PROPERTY(qreal period READ period WRITE setPeriod)
public:
    ScriptInterface(QObject *o)
        : QObject(o)
    {
        d.curve = QEasingCurve(QEasingCurve::OutElastic);
        d.curve.setOvershoot(4);
        d.duration = 1000;
    }

    QString easingCurve() const
    {
        const QMetaEnum e = easingCurveTypeEnum();
        return QString::fromLatin1(e.valueToKey(d.curve.type()));
    }

    bool setEasingCurve(const QString &type)
    {
        const QMetaEnum e = easingCurveTypeEnum();
        const int count = e.keyCount();
        for (int i=0; i<count; ++i) {
            const QString key = QString::fromLatin1(e.key(i));
            if (key.compare(type, Qt::CaseInsensitive) == 0) {
                d.curve.setType(static_cast<QEasingCurve::Type>(e.value(i)));
                return true;
            }
        }
        qWarning("'%s' is not a valid easing curve type", qPrintable(type));
        return false;
    }

    qreal overshoot() const
    {
        return d.curve.overshoot();
    }

    void setOvershoot(qreal overshoot)
    {
        d.curve.setOvershoot(overshoot);
    }

    qreal period() const
    {
        return d.curve.period();
    }

    void setPeriod(qreal period)
    {
        d.curve.setPeriod(period);
    }

    qreal amplitude() const
    {
        return d.curve.amplitude();
    }

    void setAmplitude(qreal amplitude)
    {
        d.curve.setAmplitude(amplitude);
    }

    QEasingCurve curve() const
    {
        return d.curve;
    }

    int duration() const
    {
        return d.duration;
    }

    void setDuration(int duration)
    {
        d.duration = duration;
    }
private:
    static QMetaEnum easingCurveTypeEnum()
    {
        const QMetaObject *object = &QEasingCurve::staticMetaObject;
        const QMetaEnum e = object->enumerator(object->indexOfEnumerator("Type"));
        return e;
    }
    struct Data {
        QEasingCurve curve;
        int duration;
    } d;
};

class CoverFlowManager : public QObject
{
    Q_OBJECT
public:
    CoverFlowManager(int side, GraphicsScene *scene);
    ~CoverFlowManager();
    void keyPressEvent(QKeyEvent *e);
    void setCurrentIndex(int idx);
    int currentIndex() const { return d.current; }
    QParallelAnimationGroup *createAnimationTo(int idx);
    void animateTo(int idx);
    QSize itemSize() const;
public slots:
    void updateZValues();
private:
    QRectF centerRect() const;
    QPair<QPolygonF, QPolygonF> initPolygons(int idx) const;
    qreal stackingWidth() const;
    struct Data {
        GraphicsScene *scene;
        QWeakPointer<QParallelAnimationGroup> animationGroup;
        QList<ReflectionItem*> reflections;
        int current;
        QSize itemSize;
        mutable QHash<int, QPair<QPolygonF, QPolygonF> > cachedPolygons;
    } d;
};


GraphicsScene::GraphicsScene(GraphicsView *parent)
    : QGraphicsScene(parent)
{
    d.hasNonViewItems = false;
    d.coverFlowManager = 0;
    Settings settings;
    Q_ASSERT(!inst);
    inst = this;
#if defined(DEMO_GL_FBO) || defined(DEMO_GLES_FBO)
    setBackgroundBrush(QColor(255*0.3, 255*0.3, 255*0.7));
#else
    setBackgroundBrush(Qt::black);
#endif
    d.directoryTextVisible = true;
    d.sendingDemoModeEvent = false;
    d.dotDot = 0;
    d.replayEventIdx = 0;
    d.idleTimer.start(QCoreApplication::instance()->property("idleTimer").toInt(), this);
//    d.replayEvents = qVariantValue<QList<QKeyEvent*> >(QCoreApplication::instance()->property("replayEvents"));
    d.selector = 0;
    d.frames = 0;
    d.view = parent;
    d.currentLayoutMode = LayoutModeCount;
    d.nextZOrder = 100000;
    d.verticalScrollBarAnimation = new QPropertyAnimation(parent->verticalScrollBar(), "value", this);
    connect(parent->horizontalScrollBar(), SIGNAL(valueChanged(int)), this, SLOT(updateDemoModePosition()));
    connect(parent->verticalScrollBar(), SIGNAL(valueChanged(int)), this, SLOT(updateDemoModePosition()));
    d.demoModeDuration = 2000;
    d.demoModeItem = new TextItem("Demo");
    addItem(d.demoModeItem);
    d.demoModeItem->setBrush(Qt::white);

    QAction *menuAction = new QAction("Menu", parent);
    menuAction->setSoftKeyRole(QAction::PositiveSoftKey);
    parent->addAction(menuAction);

    d.menu = new QMenu;
    connect(menuAction, SIGNAL(triggered()), this, SLOT(showMenu()));

    QAction *up = new QAction("Up", parent);
    connect(up, SIGNAL(triggered()), this, SLOT(goParentDirectory()));

    QAction *zoomOut = new QAction("Zoom out", parent);
    connect(zoomOut, SIGNAL(triggered()), parent, SLOT(zoomOut()));

    QAction *zoomIn = new QAction("Zoom in", parent);
    connect(zoomIn, SIGNAL(triggered()), parent, SLOT(zoomIn()));

    d.menu->addAction(up);
    d.menu->addAction(zoomOut);
    d.menu->addAction(zoomIn);


    if (!settings.value("nofps", false).toBool()) {
        d.fpsItem = new TextItem;
        d.fpsItem->setPos(settings.value("fpsItemX", 0).toInt(), settings.value("fpsItemY", 0).toInt());
        d.fpsItem->setBackground(Qt::black);
        d.fpsItem->setBrush(Qt::white);

        QFont font;
        Q_ASSERT(parent);
        font.setPixelSize(qMax(10, d.view->viewport()->height() / 10));
        d.fpsItem->setFont(font);
        d.fpsItem->setFlag(QGraphicsItem::ItemIgnoresTransformations);
        d.fpsItem->setOpacity(0);
        addItem(d.fpsItem);
        d.fpsItem->setZValue(SpecialItemZ);
    } else {
        d.fpsItem = 0;
    }

    d.demoModeItem->setZValue(SpecialItemZ);
    QFont font;
    font.setPixelSize(80);
    d.demoModeItem->setFont(font);
    d.demoModeItem->setOpacity(0);
    d.scriptInterface = new ScriptInterface(this);
    d.demoMode = false;
    const QString fn = QCoreApplication::instance()->property("record").toString();
    if (!fn.isEmpty()) {
        d.record = new QFile(fn);
        d.recordTime.start();
        if (!d.record->open(QIODevice::WriteOnly)) {
            delete d.record;
            qDebug("Couldn't open %s for writing", qPrintable(fn));
            d.record = 0;
        } else {
            d.record->write("__Magic__\n");
        }
    } else {
        d.record = 0;
        const QString replayFile = QCoreApplication::instance()->property("replayFile").toString();
        QFile file(replayFile);
        if (file.exists() && file.open(QIODevice::ReadOnly)) {
            QTextStream ts(&file);
            QString tmp;
            ts >> tmp;
            if (tmp != "__Magic__") {
                printf("%s %d: if (line != \"__Magic__\") {\n", __FILE__, __LINE__);
                return;
            }
            forever {
                Event event;
                ts >> event.key >> event.modifiers >> event.msecs;
                if (ts.status() != QTextStream::Ok)
                    break;
                d.replayEvents.append(event);
            }
        } else if (!replayFile.isEmpty()) {
            qWarning("Can't open %s for reading", qPrintable(replayFile));
        }
    }
    if (settings.value("startInDemoMode", false).toBool()) {
        QTimer::singleShot(3000, this, SLOT(startDemoMode()));
    }

#ifdef DEMO_GL
    setItemIndexMethod(QGraphicsScene::NoIndex);
#endif
}

GraphicsScene::~GraphicsScene()
{
    Q_ASSERT(inst == this);
    inst = 0;
    delete d.dotDot;
    delete d.record;
    if (d.imageLoadingThread.data()) {
        d.imageLoadingThread.data()->resume();
        d.imageLoadingThread.data()->abort();
        d.imageLoadingThread.data()->wait();
    }
    delete d.menu;
}


QDebug &operator<<(QDebug &dbg, const QFileInfo &fileInfo)
{
    dbg.maybeSpace() << fileInfo.absoluteFilePath();
    return dbg;
}

static inline QBitArray transitions(const QStringList &list)
{
    QBitArray ret;
    if (list.isEmpty()) {
        ret.fill(true, GraphicsScene::LayoutModeCount);
    } else {
        ret.fill(false, GraphicsScene::LayoutModeCount);
        const QMetaEnum enumerator = GraphicsScene::staticMetaObject.enumerator(GraphicsScene::staticMetaObject.indexOfEnumerator("LayoutMode"));
        Q_ASSERT(enumerator.isValid());
        foreach(const QString &transition, list) {
            const int idx = enumerator.keyToValue(transition.toLatin1().constData());
            if (idx < 0) {
                qWarning("This is not a valid mode %s", qPrintable(transition));
                continue;
            }
            ret.setBit(enumerator.value(idx), true);
        }
    }
    return ret;
}

template <typename T>
static inline void shuffle(QList<T> *list)
{
    for (int i=0; i<list->size(); ++i) {
        int idx = rand() % list->size();
        if (idx != i)
            qSwap((*list)[idx], (*list)[i]);
    }
}

#ifdef DEMO_QML
static inline QmlItem *createQmlItem(const QFileInfo &file)
{
    Q_ASSERT(file.isDir());
    QFileInfo fi(file.absoluteFilePath() + "/" + file.baseName() + ".qml");
    if (!fi.exists()) {
        const Settings settings(file.absoluteFilePath() + "/settings.conf");
        const QString entryPoint = settings.value("qmlfile").toString();
        if (!entryPoint.isEmpty()) {
            fi = QFileInfo(file.absoluteFilePath() + "/" + entryPoint);
        }
    }
    if (fi.exists()) {
        return new QmlItem(file, QUrl::fromLocalFile(fi.absoluteFilePath()));
    }

    return 0;
}
#endif

void GraphicsScene::setDirectory(const QDir &dd, LayoutMode layout, const QString &newCurrent)
{
    delete d.coverFlowManager;
    d.coverFlowManager = 0;
    d.hasNonViewItems = false;
    if (!d.selector) {
        d.selector = new SelectorItem;
        d.selector->setZValue(AlmostSpecialItemZ);
        addItem(d.selector);
    }

    Settings settings(dd.absolutePath() + "/settings.conf");
    if (settings.value("hidecursor").toBool() != bool(QApplication::overrideCursor())) {
        if (QApplication::overrideCursor()) {
            QApplication::restoreOverrideCursor();
        } else {
            QApplication::setOverrideCursor(Qt::BlankCursor);
        }
    }
    static const int max = QCoreApplication::instance()->property("maxItems").toInt();
    Q_ASSERT(max > 0);
    d.nextZOrder = 100000;

    QDir dir = dd;
    dir.setNameFilters(settings.value("nameFilters").toStringList());

    d.supportedTransitions = ::transitions(settings.value("supportedTransitions").toString().split(",", QString::SkipEmptyParts));

    if (d.imageLoadingThread.data()) {
        if (d.imageLoadingThread.data()->isRunning()) {
            d.imageLoadingThread.data()->abort();
        } else {
            delete d.imageLoadingThread.data();
        }
        d.imageLoadingThread.clear();
    }
    if (d.root.isEmpty()) {
        d.root = dir.absolutePath();
    }
    for (int i=0; i<d.items.size(); ++i) {
        QGraphicsWidget *&w = d.items[i];
        if (w->property("itemType").toInt() == WebPageType) {
            removeItem(w);
            w = 0;
        }
    }

    qDeleteAll(d.items);
    d.items.clear();
    d.pendingImageItems.clear();
    QLineF line(availableScreenRect().center(), QPointF(1, 1));
    line.setLength(availableScreenRect().width() * 1.5);
    QList<QFileInfo> fileInfos = dir.entryInfoList(QDir::Files|QDir::Dirs|QDir::NoDotAndDotDot);
    if (max < fileInfos.size()) {
        shuffle(&fileInfos);
    }
    const QSizeF is = itemSize();
    delete d.dotDot;
    if (dir.absolutePath() != d.root) {
        d.dotDot = new QDir(dir);
        d.dotDot->cdUp();
    } else {
        d.dotDot = 0;
    }
    d.onlySelectedTextVisible = settings.value("onlySelectedTextVisible", false).toBool();
    d.directoryTextVisible = settings.value("textVisible", true).toBool();

    if (!d.directoryTextVisible) {
        d.onlySelectedTextVisible = false;
    }

    if (d.onlySelectedTextVisible) {
        d.directoryTextVisible = false;
    }

    QList<QPair<QString, void*> > pendingImages;
    bool isWeb = false;

    QHash<QString, QString> scriptForFiles;
    const QList<QByteArray> imageExtensions = QImageReader::supportedImageFormats();
    int current = 0;
    for (int i=0; i<fileInfos.size() && d.items.size() < max; ++i) {
        const QFileInfo &fileInfo = fileInfos.at(i);
        const QString fn = fileInfo.fileName();
        if (fn.endsWith("_full.png")) // these are used for setting backgrounds and when zooming in
            continue;
        const QByteArray extension = fileInfo.suffix().toLatin1().toLower();
        if (extension == "js") {
            QFile file(fileInfo.absoluteFilePath());
            file.open(QIODevice::ReadOnly);
            const QString script = QString::fromLocal8Bit(file.readAll());
            if (fn == "all.js") {
                scriptForFiles[QString()] = script;
            } else {
                scriptForFiles[fn.left(fn.size() - 3)] = script;
            }
            continue;
        } else if (extension == "ignore") {
            continue;
        } else if (fn == "dir.png") {
            continue;
        } else if (fn == "settings.conf") {
            continue;
        }
        QGraphicsWidget *item;
        if (imageExtensions.contains(extension)) {
            ImageItem *it = new ImageItem(fileInfo);
            it->setText(fileInfo.baseName());
            d.pendingImageItems.insert(it);
            pendingImages.append(qMakePair<QString, void*>(fileInfo.absoluteFilePath(), it));
            item = it;
        } else if (extension == "url") {
            WebItem *&webItem = d.cachedWebItems[fileInfo.absoluteFilePath()];
            if (!webItem) {
                webItem = new WebItem(fileInfo);
            }
            d.hasNonViewItems = true;
            webItem->page()->mainFrame()->setScrollBarValue(Qt::Vertical, 0);
            item = webItem;
            isWeb = true;
        } else if (fileInfo.isDir()) {
            Settings dirSettings(fileInfo.absoluteFilePath() + "/settings.conf");
            if (dirSettings.value("hidden").toBool() == true) {
                continue;
            }
#ifdef DEMO_QML
            item = ::createQmlItem(fileInfo);
#else
            if (fileInfo.fileName() == "Qml")
                continue;
            item = 0;
#endif
            if (!item) {
                item = new DirectoryItem(fileInfo);
            }
            const QString name = dirSettings.value("name").toString();
            if (!name.isEmpty())
                item->setProperty("text", name);
        } else if (extension == "so") {
            PluginItem *plugin = new PluginItem(fileInfo);
            if (!plugin->isValid()) {
                delete plugin;
                continue;
            }
            item = plugin;
        } else {
            item = new FileItem(fileInfo);
        }
        if (!newCurrent.isEmpty() && fileInfo.absoluteFilePath().contains(newCurrent)) {
            current = d.items.size();
        }
        item->resize(is);
        if (!d.directoryTextVisible || d.onlySelectedTextVisible) {
            item->setProperty("textVisible", false);
        }
        line.setAngle(rand() % 360);
        d.items.append(item);
        item->setPos(line.p2());
        addItem(item);
    }
    if (isWeb) {
        const QRectF sr = availableScreenRect();
        const QSizeF size((sr.width() / d.items.size()) - 15, sr.height());
        foreach(QGraphicsWidget *widget, d.items) {
            widget->resize(size);
            Q_ASSERT(widget->size() == size);
        }
    }

    if (!scriptForFiles.isEmpty()) {
        int i = 0;
        while (i < d.items.size()) {
            QGraphicsWidget *item = d.items.at(i++);
            const QString file = qVariantValue<QFileInfo>(item->property("fileInfo")).fileName();
            if (scriptForFiles.contains(file)) {
                item->setProperty("script", scriptForFiles.value(file));
            } else {
                item->setProperty("script", scriptForFiles.value(QString()));
            }
        }
    }
    updateDemoModePosition();
    if (!pendingImages.isEmpty()) {
        d.imageLoadingThread = new ImageLoadingThread(QCoreApplication::instance()->property("--image-throttle").toInt(),
                                                      itemSize());
        d.imageLoadingThread.data()->setImages(pendingImages);
        connect(d.imageLoadingThread.data(), SIGNAL(imageLoaded(QImage,void*)),
                this, SLOT(onImageLoaded(QImage,void*)));
    }

    d.currentLayoutMode = LayoutModeCount;
    if (!d.items.isEmpty())
        setCurrentItem(d.items.at(current));
    if (layout != LayoutModeCount) {
        setLayout(layout);
        if (d.lastAnimation && settings.value("notransition", false).toBool()) {
            d.lastAnimation.data()->setCurrentTime(d.lastAnimation.data()->totalDuration());
        }
    }
    d.current = dir;
    d.selector->setVisible(currentItem());
    
}

void GraphicsScene::showMenu()
{
    d.menu->exec();
}

void GraphicsScene::goParentDirectory()
{
    if (d.dotDot) {
        const QString path = d.current.absolutePath();
        const QDir dir = *d.dotDot;
        setDirectory(dir, Grid, path);
    }
}

QSizeF GraphicsScene::itemSize() const
{
    if (QGraphicsWidget *first = d.items.value(0))
        return first->size();
    static const int itemSize = QCoreApplication::instance()->property("itemSize").toInt();
    return QSizeF(itemSize, itemSize);
}

void GraphicsScene::restoreCurrentItem()
{
    if (d.currentLayoutMode == ZoomedAndIsolated) {
        currentItem()->setProperty("fillBackground", false);
        QPointF pos = currentItem()->property("original_position_beforezoom").toPointF();
        QSizeF size = currentItem()->property("original_size_beforezoom").toSizeF();
#ifdef ROTATIONS_ENABLED
        currentItem()->setTransform(QTransform());
#endif
        currentItem()->setGeometry(pos.x(), pos.y(), size.width(), size.height());
        QMetaObject::invokeMethod(currentItem(), "showSmallPixmap");
        d.selector->show();
        d.selector->adjustGeometry();
    }
}

static GraphicsScene::Direction direction(const QKeyEvent *e)
{
    if (!e->modifiers()) {
        switch (e->key()) {
        case Qt::Key_Left: return GraphicsScene::Left;
        case Qt::Key_Right: return GraphicsScene::Right;
        case Qt::Key_Up: return GraphicsScene::Up;
        case Qt::Key_Down: return GraphicsScene::Down;
        }
    }
    return GraphicsScene::None;
}

void GraphicsScene::keyPressEvent(QKeyEvent *e)
{
    d.idleTimer.start(QCoreApplication::instance()->property("idleTimer").toInt(), this);
    if (activePanel()) {
        QGraphicsScene::keyPressEvent(e);
        return;
    }
    e->accept();
    if (d.progressItem) {
        return;
    }
    if (d.imageLoadingThread.data()) {
        d.imageLoadingThread.data()->suspend();
        d.suspendThreadTimer.start(1000, this);
    }
    if (d.record) {
        QTextStream ts(d.record);
        ts << e->key() << " " << e->modifiers() << " " << d.recordTime.restart() << endl;
    }
    if (d.demoMode && !d.sendingDemoModeEvent) {
        if (e->key() == Qt::Key_R)
            restartDemoMode();
        else
            stopDemoMode();
        return;
    }

    if (d.coverFlowManager) {
        d.coverFlowManager->keyPressEvent(e);
        if (e->isAccepted())
            return;
    }

    // handle mode agnostic keys here.
    if (e->modifiers() == Qt::ControlModifier)
        switch (e->key()) {
        case Qt::Key_D:
            startDemoMode();
            break;
        case Qt::Key_L:
            if (qobject_cast<QGraphicsWebView *>(d.items.at(0))) {
                queryUrlDialog();
                return;
            }
            break;
        case Qt::Key_B:
            if (currentItem()
                && currentItem()->property("itemType").toInt() == ImageType) {
                ImageItem *img = static_cast<ImageItem*>(currentItem());
                QString fileName = img->fileInfo().absoluteFilePath();
                Settings settings(img->fileInfo().absolutePath() + "/settings.conf");
                if (!settings.value("wallpapers").toBool() || !settings.value("background").toBool())
                    return;
                fileName.replace("_thumbnail", "_full");
                QCoreApplication::instance()->setProperty("backgroundImage", fileName);
                update();
                return;
            }
            break;
        case Qt::Key_U:
            if (currentItem()
                && currentItem()->property("itemType").toInt() == QmlType) {
                extern void showSource(const QFileInfo &);
                showSource(qVariantValue<QFileInfo>(currentItem()->property("fileInfo")));
                return;
            }
            break;
        case Qt::Key_F:
            if (d.fpsItem) {
                d.fpsItem->setVisible(!d.fpsItem->isVisible());
                return;
            }
        default:
            break;
        }

    int mode = layoutModeForKey(e->key(), e->modifiers());
    if (d.currentLayoutMode == ZoomedAndIsolated) {
        if (mode < LayoutModeCount
#ifdef FRIENDLY_KEYS
            || e->key() == Qt::Key_Escape
#endif
            || ((e->modifiers() & Qt::ControlModifier) && e->key() == Qt::Key_Left)) {
            mode = ZoomedAndIsolated;
        } else {
            QGraphicsScene::keyPressEvent(e);
            e->accept();
            return;
        }
    }
    if (!currentItem() && !d.items.isEmpty()) {
        d.items.first()->setFocus();
        return;
    }

    if (mode < LayoutModeCount) {
        setLayout(static_cast<LayoutMode>(mode));
    } else {
        const Direction direction = ::direction(e);
        if (direction != None && currentItem()) {
            QGraphicsWidget *newCurrent = findItem(direction);
            if (newCurrent && newCurrent != currentItem()) {
                setCurrentItem(newCurrent);
                centerOn(newCurrent);
            }
            return;
        }

        switch (e->key()) {
        case Qt::Key_Right:
            if (e->modifiers() != Qt::ControlModifier)
                break;
            // fallthrough
#ifdef FRIENDLY_KEYS
        case Qt::Key_Return:
        case Qt::Key_Enter:
        case Qt::Key_Select:
#endif
            if (currentItem()) {
                activateCurrentItem();
                return;
            }
            break;
        case Qt::Key_Left:
            if (e->modifiers() != Qt::ControlModifier)
                break;
            // fallthrough
#ifdef FRIENDLY_KEYS
        case Qt::Key_Escape:
        case Qt::Key_Backspace:
#endif
            goParentDirectory();
            return;
            break;
        default:
            break;
        }
    }
    e->ignore();
    QGraphicsScene::keyPressEvent(e);
}



void GraphicsScene::setCurrentItem(QGraphicsWidget *newCurrent)
{
    if (currentItem()) {
        Q_ASSERT(currentItem()->property("_oldz").isValid());
        currentItem()->setZValue(currentItem()->property("_oldz").toInt());
        if (d.onlySelectedTextVisible)
            currentItem()->setProperty("textVisible", false);
    }
    if (newCurrent) {
        setFocusItem(newCurrent);
        Q_ASSERT(newCurrent == focusItem());
        Q_ASSERT(newCurrent == currentItem());

        Q_ASSERT(newCurrent->flags() & QGraphicsItem::ItemIsFocusable);
        currentItem()->setProperty("_oldz", currentItem()->zValue());
        currentItem()->setZValue(CurrentItemZ);
        if (d.onlySelectedTextVisible) {
            currentItem()->setProperty("textVisible", true);
        }
    }
    if (!d.coverFlowManager)
        d.selector->select(currentItem());
}

QAbstractAnimation * GraphicsScene::createEllipse() const
{
    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    QRectF sr = availableScreenRect();
    const QSize is = itemSize().toSize();
    sr.setHeight(sr.height() - is.height());
    sr.moveTop(is.height() / 2);
    const qreal radius = qMin(sr.width(), sr.height()) / 2;
    for (int i=0; i<d.items.size(); ++i) {
        if (d.items.at(i) != currentItem()) {
            d.items.at(i)->setZValue(i);
        }
        QLineF line(sr.center(), sr.center() + QPointF(radius, 0));
        line.setAngle((360.0 / qreal(d.items.size())) * qreal(i));
        const QPointF pp = line.p2() - QPointF(is.width() / 2, is.height() / 2);
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "pos");
        animation->setEasingCurve(curve());
        animation->setEndValue(pp);
        animation->setDuration(duration());
        group->addAnimation(animation);
    }
    return group;
}


QAbstractAnimation * GraphicsScene::createShuffle() const
{
    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    const QSize sr = availableScreenRect().size().toSize();
    const QSize is = itemSize().toSize();
    Q_ASSERT(is.width() < sr.width() && is.height() < sr.height());
    for (int i=0; i<d.items.size(); ++i) {
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "pos");
        animation->setEasingCurve(curve());

        animation->setEndValue(QPointF(qrand() % (sr.width() - is.width()),
                                       qrand() % (sr.height() - is.height())));
        animation->setDuration(duration());
        group->addAnimation(animation);
    }
    return group;
}

QAbstractAnimation * GraphicsScene::createGrid() const
{
    const qreal width = availableScreenRect().width();
    enum { MinimumMargin = 10 };
    const int columns = width / (itemSize().width() + MinimumMargin);

    Q_ASSERT(columns > 0);
    const QSizeF item = itemSize();
    const qreal margin = (width - (columns * item.width())) / (columns + 1);

    const QSizeF cell(item.width() + margin, item.height() + margin);

    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    QPointF p = QPointF(margin, margin) - QPointF(cell.width(), 0);
    for (int i=0; i<d.items.size(); ++i) {
        if (d.items.at(i) != currentItem()) {
            d.items.at(i)->setZValue(i);
        }
        if (i > 0 && i % columns == 0) {
            p.rx() = margin;
            p.ry() += cell.height();
        } else {
            p.rx() += cell.width();
        }
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "pos");
        animation->setEasingCurve(curve());
        animation->setEndValue(p);
        animation->setDuration(duration());
        group->addAnimation(animation);
    }

    return group;
}

void GraphicsScene::drawBackground(QPainter *p, const QRectF &rr)
{
    const QString appBackground = QCoreApplication::instance()->property("backgroundImage").toString();
    if (d.items.isEmpty() || appBackground.isEmpty()) {
        QGraphicsScene::drawBackground(p, rr);
        return;
    }
    if (appBackground != d.backgroundFileName) {
        d.backgroundFileName = appBackground;
        d.backgroundPixmap = ::loadPixmap(appBackground);
    }
    const QTransform old = p->transform();
    QPoint pp;
    QRectF r = rr;
    if (old.isTranslating()) {
        r = p->transform().mapRect(rr);
        p->setTransform(QTransform());
    }
    const QRect viewRect = d.view->viewport()->rect();
    if (viewRect.size() != d.backgroundPixmap.size())
        p->fillRect(r, backgroundBrush());
    QRectF pixRect = d.backgroundPixmap.rect();
    pixRect.moveCenter(viewRect.center());
    const QPointF offset = pixRect.topLeft();
    pixRect &= r;
    p->drawPixmap(pixRect, d.backgroundPixmap, pixRect.translated(-offset));
    if (old.isTranslating())
        p->setTransform(old);
}

void GraphicsScene::drawForeground(QPainter *painter, const QRectF &rect)
{
    ++d.frames;
    QGraphicsScene::drawForeground(painter, rect);
}

static inline qreal distance(const QPointF &from, const QPointF &to, Qt::Orientation orientation)
{
    qreal len = QLineF(from, to).length();
    len += 4 * (orientation == Qt::Horizontal ? qAbs(from.y() - to.y()) : qAbs(from.x() - to.x()));
    return len;
}

QGraphicsWidget * GraphicsScene::findItem(Direction direction) const
{
    const QList<QGraphicsItem *> i = items();
    QGraphicsItem *best = 0;
    qreal bestDistance = INT_MAX;
    Q_ASSERT(currentItem());
    const QPointF pos = currentItem()->pos();
    foreach(QGraphicsItem *item, i) {
        if (item == currentItem() || !(item->flags() & QGraphicsItem::ItemIsFocusable))
            continue;
        const QRectF rect = item->sceneBoundingRect();
        Qt::Orientation orientation = Qt::Horizontal;
        switch (direction) {
        case None: Q_ASSERT(0); break;
        case Left:
            if (rect.x() >= pos.x() || qFuzzyCompare(rect.x(), pos.x()))
                continue;
            orientation = Qt::Horizontal;
            break;
        case Right:
            if (rect.x() <= pos.x() || qFuzzyCompare(rect.x(), pos.x()))
                continue;
            orientation = Qt::Horizontal;
            break;
        case Up:
            if (rect.y() >= pos.y() || qFuzzyCompare(rect.y(), pos.y()))
                continue;
            orientation = Qt::Vertical;
            break;
        case Down:
            if (rect.y() <= pos.y() || qFuzzyCompare(rect.y(), pos.y()))
                continue;
            orientation = Qt::Vertical;
            break;
        }
        const qreal dist = distance(pos, rect.topLeft(), orientation);
        if (!best || dist < bestDistance) {
            bestDistance = dist;
            best = item;
        }
    }
    return static_cast<QGraphicsWidget*>(best);
}

QAbstractAnimation * GraphicsScene::createDiagonal() const
{
    if (d.items.size() <= 1)
        return 0;
    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    const QSizeF is = itemSize();
    QRectF sr = availableScreenRect();
    const QSizeF vs = sr.size() - is;
    const QPointF add(vs.width() / (d.items.size() - 1),
                      vs.height() / (d.items.size() - 1));
    QPointF p(sr.left(), sr.top());
    for (int i=0; i<d.items.size(); ++i) {
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "pos");
        animation->setEasingCurve(curve());
        if (d.items.at(i) != currentItem()) {
            d.items.at(i)->setZValue(i);
        }
        animation->setEndValue(p);
        p += add;
        animation->setDuration(duration());
        group->addAnimation(animation);
    }
    return group;
}

QAbstractAnimation * GraphicsScene::createDiagonal2() const
{
    if (d.items.size() <= 1)
        return 0;

    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    const QSizeF is = itemSize();
    QRectF sr = availableScreenRect();
    const QSizeF vs = sr.size() - is;
    const QPointF add(-vs.width() / (d.items.size() - 1),
                      vs.height() / (d.items.size() - 1));
    QPointF p(vs.width(), sr.top());
    for (int i=0; i<d.items.size(); ++i) {
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "pos");
        animation->setEasingCurve(curve());
        if (d.items.at(i) != currentItem()) {
            d.items.at(i)->setZValue(i);
        }
        animation->setEndValue(p);
        p += add;
        animation->setDuration(duration());
        group->addAnimation(animation);
    }
    return group;
}

void GraphicsScene::timerEvent(QTimerEvent *e)
{
    if (e->timerId() == d.demoTimer.timerId()) {
        Q_ASSERT(d.replayEventIdx < d.replayEvents.size());
        if (d.replayEventIdx < d.replayEvents.size()) {
            const Event &event = d.replayEvents.at(d.replayEventIdx);
            QKeyEvent e(QEvent::KeyPress, event.key, (Qt::KeyboardModifiers)event.modifiers);
            d.sendingDemoModeEvent = true;
            QCoreApplication::sendEvent(this, &e);
            d.sendingDemoModeEvent = false;
            if (d.replayEventIdx + 1 < d.replayEvents.size()) {
                d.demoTimer.start(d.replayEvents.at(d.replayEventIdx + 1).msecs, this);
                ++d.replayEventIdx;
            } else {
                restartDemoMode();
            }
        }
    } else if (e->timerId() == d.suspendThreadTimer.timerId()) {
        if (d.imageLoadingThread.data()) {
            d.imageLoadingThread.data()->resume();
        }
        d.suspendThreadTimer.stop();
    } else if (e->timerId() == d.idleTimer.timerId()) {
        if (stack()->property("allowDemoMode").toBool())
            startDemoMode();
        d.idleTimer.stop();
    } else {
        QGraphicsScene::timerEvent(e);
    }
}

static inline qreal space(qreal length, int itemCount)
{
    return length / itemCount;
}

QAbstractAnimation * GraphicsScene::createOutline() const
{
    if (d.items.isEmpty())
        return 0;

    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    const QRectF sr = availableScreenRect();
    const QSizeF is = itemSize();
    const qreal horizontal = sr.width() - (is.width() * 2);
    const qreal vertical = sr.height() - (is.height() * 2);
    int items = d.items.size() - 4;
    const qreal horizontalRatio = horizontal / (horizontal + vertical);
    const qreal verticalRatio = vertical / (vertical + horizontal);
    int horizontalTop = qRound(items * (horizontalRatio / 2));
    int verticals = qRound(items * (verticalRatio / 2));
    int horizontalBottom = items - horizontalTop - (verticals * 2);
    if (horizontalBottom > horizontalTop)
        qSwap(horizontalTop, horizontalBottom);
    int idx = 0;
    QPropertyAnimation *animation = 0;
    QPointF pos = sr.topLeft();
    qreal space = ::space(sr.width() - is.width(), horizontalTop + 1);
    for (int i=0; i<horizontalTop + 2; ++i) {
        QGraphicsWidget *item = d.items.at(idx);
        animation = new QPropertyAnimation(item, "pos");
        animation->setDuration(duration());
        animation->setEasingCurve(curve());
        animation->setEndValue(pos);
        if (item != currentItem()) {
            item->setZValue(idx);
        }
        group->addAnimation(animation);
        ++idx;
        pos += QPointF(space, 0);
    }

    pos = sr.topRight() - QPointF(is.width(), 0);
    space = ::space(sr.height() - (2 * is.height()), verticals);
    for (int i=0; i<verticals; ++i) {
        pos += QPointF(0, space);
        QGraphicsWidget *item = d.items.at(idx);
        animation = new QPropertyAnimation(item, "pos");
        animation->setEndValue(pos);
        animation->setDuration(duration());
        animation->setEasingCurve(curve());
        if (item != currentItem()) {
            item->setZValue(idx);
        }
        group->addAnimation(animation);
        ++idx;
    }

    pos = sr.bottomLeft() - QPointF(0, is.height());
    space = ::space(sr.width() - is.width(), horizontalBottom + 1);
    for (int i=0; i<horizontalBottom + 2; ++i) {
        QGraphicsWidget *item = d.items.at(idx);
        animation = new QPropertyAnimation(item, "pos");
        animation->setDuration(duration());
        animation->setEasingCurve(curve());
        animation->setEndValue(pos);
        if (item != currentItem()) {
            item->setZValue(idx);
        }
        group->addAnimation(animation);
        ++idx;
        pos += QPointF(space, 0);
    }

    pos = sr.topLeft();
    space = ::space(sr.height() - (2 * is.height()), verticals);
    for (int i=0; i<verticals; ++i) {
        pos += QPointF(0, space);
        QGraphicsWidget *item = d.items.at(idx);
        animation = new QPropertyAnimation(item, "pos");
        animation->setDuration(duration());
        animation->setEasingCurve(curve());
        animation->setEndValue(pos);
        if (item != currentItem()) {
            item->setZValue(idx);
        }
        group->addAnimation(animation);
        ++idx;
    }
    return group;
}

int GraphicsScene::duration() const
{
    return d.scriptInterface->duration();
}

QEasingCurve GraphicsScene::curve() const
{
    return d.scriptInterface->curve();
}

void GraphicsScene::activate(QGraphicsWidget *widget)
{
    Q_ASSERT(widget);
    const QString script = widget->property("script").toString();
    if (!script.isEmpty()) {
        QScriptEngine engine;
        engine.globalObject().setProperty("animation",
                                          engine.newQObject(d.scriptInterface, QScriptEngine::QtOwnership,
                                                            QScriptEngine::ExcludeSuperClassContents));
        QFileInfo fileInfo = qVariantValue<QFileInfo>(widget->property("fileInfo"));
        engine.globalObject().setProperty("absoluteFilePath", fileInfo.absoluteFilePath());
        engine.globalObject().setProperty("fileName", fileInfo.fileName());
        QScriptValue val = engine.evaluate(script);
        if (engine.hasUncaughtException()) {
            qWarning("%s", qPrintable(engine.uncaughtException().toString()));
        }
        return;
    }
    const QVariant type = widget->property("itemType");
    if (type.isValid()) {
        switch (type.toInt()) {
        case FileType:
            break;
        case DirectoryType:
            setDirectory(qVariantValue<QDir>(widget->property("dir")).absolutePath());
            break;
        case ImageType:
        case WebPageType:
            setLayout(ZoomedAndIsolated);
            break;
        case VideoType:
            break;
        case PluginType:
            static_cast<PluginItem*>(widget)->exec();
            break;
#ifdef DEMO_QML
        case QmlType:
            static_cast<QmlItem*>(widget)->exec();
            break;
#endif
        }
    }
}

void GraphicsScene::startDemoMode()
{
    if (d.replayEvents.isEmpty()) {
        return;
    }
    if (d.dotDot) {
        setDirectory(QDir("./demo"));
    }
    d.replayEventIdx = 0;
    Q_ASSERT(!d.demoMode);
    d.demoMode = true;
    d.demoTimer.start(d.replayEvents.first().msecs, this);
    QPropertyAnimation *animation = new QPropertyAnimation(d.demoModeItem, "opacity", d.demoModeItem);
    animation->setEndValue(1.0);
    animation->setDuration(duration());
    animation->start(QAbstractAnimation::DeleteWhenStopped);
}

void GraphicsScene::stopDemoMode()
{
    if (!d.demoMode) {
        return;
    }
    d.demoMode = false;
    d.demoTimer.stop();
    QPropertyAnimation *animation = new QPropertyAnimation(d.demoModeItem, "opacity", d.demoModeItem);
    animation->setEndValue(0.0);
    animation->setDuration(duration());
    animation->start(QAbstractAnimation::DeleteWhenStopped);
}

void GraphicsScene::restartDemoMode()
{
    if (!d.demoMode) {
        return;
    }
    d.demoMode = false;
    d.demoTimer.stop();

    if (d.dotDot) {
        setDirectory(QDir("./demo"));
    }
    setCurrentItem(d.items.at(0));

    QTimer::singleShot(100, this, SLOT(startDemoMode()));
}

void GraphicsScene::onImageLoaded(const QImage &image, void *ptr)
{
    ImageItem *item = static_cast<ImageItem*>(ptr);
    if (d.pendingImageItems.remove(item)) {
        item->setPixmap(QPixmap::fromImage(image));
    }
}

void GraphicsScene::setLayout(LayoutMode mode)
{
    if (!currentItem() || mode >= LayoutModeCount
#ifdef ROTATIONS_ENABLED
        || (d.hasNonViewItems && mode == CoverFlow)
#endif
        || !d.supportedTransitions.testBit(mode)) {
        return;
    }

    if (d.imageLoadingThread.data()) {
        d.imageLoadingThread.data()->suspend();
        d.suspendThreadTimer.start(duration() + 100, this);
    }
    d.pendingViewportUpdateMode = d.view->viewportUpdateMode();
    d.view->setViewportUpdateMode(QGraphicsView::FullViewportUpdate);

    restoreCurrentItem();
    if (d.fpsItem) {
        delete d.lastAnimation.data();
        d.frames = 0;
        d.started = QTime::currentTime();
        delete qFindChild<QSequentialAnimationGroup*>(d.fpsItem);
    }

    if (mode != ZoomedAndIsolated)
        setSceneRect(d.view->viewport()->rect());

    QAbstractAnimation *animation = 0;
    switch (mode) {
    case ZoomedAndIsolated:
        if (d.currentLayoutMode != ZoomedAndIsolated) {
            animation = createZoomedAndIsolated();
            d.selector->hide();
        }
        break;
    case Shuffle: animation = createShuffle(); break;
    case Ellipse: animation = createEllipse(); break;
    case Grid: animation = createGrid(); break;
    case Outline: animation = createOutline(); break;
    case Diagonal: animation = createDiagonal(); break;
    case Diagonal2: animation = createDiagonal2(); break;
    case Collapse: animation = createCollapse(); break;
#ifdef ROTATIONS_ENABLED
    case RotateX: animation = createRotate(Qt::XAxis); break;
    case RotateY: animation = createRotate(Qt::YAxis); break;
    case RotateZ: animation = createRotate(Qt::ZAxis); break;
    case CoverFlow:
        if (!d.coverFlowManager) {
            Settings settings(d.current.absolutePath() + "/settings.conf");
            const int side = settings.value("coverflowsize", visibleSceneRect().height() / 2).toInt();
            d.coverFlowManager = new CoverFlowManager(side, this);
            d.selector->hide();
        }
        animation = d.coverFlowManager->createAnimationTo(d.coverFlowManager->currentIndex());
        break;
#endif
    case LayoutModeCount: Q_ASSERT(0); break;
    }
    if (animation) {
        d.lastAnimation = animation;
        d.lastAnimation.data()->setParent(this);
        d.currentLayoutMode = mode;
        Q_ASSERT(!d.items.isEmpty());
        connect(d.lastAnimation.data(), SIGNAL(finished()), this, SLOT(onLayoutFinished()));
        d.lastAnimation.data()->start(QAbstractAnimation::DeleteWhenStopped);
    } else {
        d.currentLayoutMode = LayoutModeCount;
    }
#ifdef ROTATIONS_ENABLED
    if (d.currentLayoutMode != CoverFlow && d.coverFlowManager) {
        delete d.coverFlowManager;
        d.coverFlowManager = 0;
        d.selector->select(currentItem());
    }
#endif
}

void GraphicsScene::onLayoutFinished()
{
    if (d.imageLoadingThread.data() && !d.imageLoadingThread.data()->isRunning()) {
        d.imageLoadingThread.data()->start();
    }
    d.pendingViewportUpdateMode = d.view->viewportUpdateMode();
    d.view->setViewportUpdateMode(d.pendingViewportUpdateMode);

    computeSceneRect();

    if (d.fpsItem) {
        const int elapsed = d.started.elapsed();
        delete d.fpsItemFadeAnimation.data();
        QSequentialAnimationGroup *group = new QSequentialAnimationGroup(d.fpsItem);
        QPropertyAnimation *fade = new QPropertyAnimation(d.fpsItem, "opacity");
        fade->setEndValue(1.0);
        fade->setDuration(200);
        group->addAnimation(fade);
        group->addPause(2500);
        fade = new QPropertyAnimation(d.fpsItem, "opacity");
        fade->setEndValue(0.0);
        fade->setDuration(200);
        group->addAnimation(fade);
        d.fpsItemFadeAnimation = group;
        group->start(QAbstractAnimation::DeleteWhenStopped);
        d.fpsItem->setText(QString("%1 fps").arg(qRound(qreal(d.frames) / qreal(elapsed / 1000.0))));
    }
    if (currentItem() && d.currentLayoutMode != ZoomedAndIsolated) {
        centerOn(currentItem());
    }
    d.selector->select(currentItem());
}

ImageLoadingThread::ImageLoadingThread(int throttle, const QSizeF &size, QObject *parent)
    : QThread(parent)
{
    connect(this, SIGNAL(finished()), this, SLOT(deleteLater()));
    d.throttle = throttle;
    d.imageSize = int(size.width());
}

void ImageLoadingThread::run()
{
    forever {
        if (isSuspended()) {
            msleep(200);
            continue;
        }
        QPair<QString, void*> pair;
        {
            QMutexLocker locker(&d.mutex);
            if (d.images.isEmpty()) {
                break;
            }
            pair = d.images.takeFirst();
        }
        QImageReader imageReader(pair.first);
        QSize s = imageReader.size();
        const QSize size(d.imageSize, d.imageSize);
        if (s != size) {
            s.scale(size, Qt::KeepAspectRatio);
            imageReader.setScaledSize(s);
        }
        const QImage image = imageReader.read();
        if (!image.isNull()) {
            emit imageLoaded(image, pair.second);
        }

        if (d.throttle)
            msleep(d.throttle);
    }
}

bool ImageLoadingThread::isSuspended() const
{
    QMutexLocker locker(&d.suspensionMutex);
    return d.suspended;
}

void ImageLoadingThread::suspend()
{
    QMutexLocker locker(&d.suspensionMutex);
    d.suspended = true;
}

void ImageLoadingThread::resume()
{
    QMutexLocker locker(&d.suspensionMutex);
    d.suspended = false;
}

void ImageLoadingThread::setImages(const QList<QPair<QString, void*> > &images)
{
    Q_ASSERT(QThread::currentThread() == thread());
    Q_ASSERT(QThread::currentThread() != this);
    QMutexLocker locker(&d.mutex);
    d.images = images;
}

void ImageLoadingThread::abort()
{
    Q_ASSERT(QThread::currentThread() == thread());
    Q_ASSERT(QThread::currentThread() != this);
    setImages(QList<QPair<QString, void*> >());
}

void GraphicsScene::updateDemoModePosition()
{
    const QRectF vr = visibleSceneRect();

    QRectF r = d.demoModeItem->sceneBoundingRect();
    r.moveTopRight(vr.topRight());
    d.demoModeItem->setPos(r.topLeft());
}

QRectF GraphicsScene::availableScreenRect() const
{
    const int margin = 5;
    QRect availRect = d.view->viewport()->rect().adjusted(margin, margin, -margin, -margin);
    return d.view->mapToScene(availRect).boundingRect();
}

QRectF GraphicsScene::visibleSceneRect() const
{
    const QRectF r = d.view->mapToScene(d.view->viewport()->rect()).boundingRect();
    return r;
}

void GraphicsScene::activateCurrentItem()
{
    activate(currentItem());
}

void GraphicsScene::centerOn(QGraphicsWidget *item)
{
    const QRectF itemRect = item->geometry();
    const QPointF pos = itemRect.center();
    QPointF viewPoint = d.view->transform().map(pos);
    d.verticalScrollBarAnimation->setEndValue(qMax(0, int(viewPoint.y() - d.view->viewport()->height() / 2.0)));
    d.verticalScrollBarAnimation->start();
}

QAbstractAnimation *  GraphicsScene::createCollapse() const
{
    if (!currentItem())
        return 0;

    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    const QRectF rect = currentItem()->geometry();
    for (int i=0; i<d.items.size(); ++i) {
        if (d.items.at(i) != currentItem()) {
            QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), "geometry");
            animation->setEasingCurve(curve());
            animation->setEndValue(rect);
            animation->setDuration(duration());
            group->addAnimation(animation);
        }
    }
    return group;
}


#ifdef ROTATIONS_ENABLED
QAbstractAnimation * GraphicsScene::createRotate(Qt::Axis axis) const
{
    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    static const char *properties[] = { "xRotation", "yRotation", "rotation", 0 };
    for (int i=0; i<d.items.size(); ++i) {
        ViewItem *viewItem = qobject_cast<ViewItem*>(d.items.at(i));
        Qt::Axis a = viewItem ? axis : Qt::ZAxis;
        QPropertyAnimation *animation = new QPropertyAnimation(d.items.at(i), properties[a]);
        animation->setEasingCurve(curve());
        animation->setEndValue(360);
        animation->setDuration(duration());
        group->addAnimation(animation);
    }
    connect(group, SIGNAL(destroyed(QObject*)), this, SLOT(resetRotations()));
    return group;
}

void GraphicsScene::resetRotations()
{
    foreach(QGraphicsWidget *widget, d.items) {
        widget->setRotation(0);
        if (ViewItem *viewItem = qobject_cast<ViewItem*>(widget)) {
            viewItem->resetTransform();
        } else {
            widget->resetTransform();
        }
    }
}

#endif


QAbstractAnimation * GraphicsScene::createZoomedAndIsolated() const
{
    currentItem()->setProperty("original_position_beforezoom", currentItem()->pos());
    currentItem()->setProperty("original_size_beforezoom", currentItem()->size());
    QSequentialAnimationGroup *sequential = new QSequentialAnimationGroup;
    QParallelAnimationGroup *group = new QParallelAnimationGroup;
    sequential->addAnimation(group);
    QPropertyAnimation *posAnimation = new QPropertyAnimation(currentItem(), "pos");
    posAnimation->setEasingCurve(curve());
    posAnimation->setDuration(duration() * 1.5);
    posAnimation->setEndValue(d.view->view().topLeft());
    group->addAnimation(posAnimation);

    QPropertyAnimation *sizeAnimation = new QPropertyAnimation(currentItem(), "size");
    sizeAnimation->setEasingCurve(curve());
    sizeAnimation->setDuration(duration() * 1.5);
    sizeAnimation->setEndValue(d.view->view().size());
    group->addAnimation(sizeAnimation);

    if (currentItem()->property("itemType").toInt() == ImageType)
        connect(group, SIGNAL(finished()), currentItem(), SLOT(showEnlarged()));

#ifdef ROTATIONS_ENABLED
    if (currentItem()->property("itemType").toInt() != WebPageType) {
        sequential->addPause(1000);
        QPropertyAnimation *rotation = new QPropertyAnimation(currentItem(), "yRotation");
        rotation->setEndValue(180);
        rotation->setDuration(duration() * 2);
        rotation->setEasingCurve(curve());
        sequential->addAnimation(rotation);
        sequential->addPause(1000);

        rotation = new QPropertyAnimation(currentItem(), "yRotation");
        rotation->setEndValue(0);
        rotation->setDuration(duration() * 2);
        rotation->setEasingCurve(curve());
        sequential->addAnimation(rotation);
    }
#endif

    currentItem()->setProperty("fillBackground", true);

    return sequential;
}

QPointF GraphicsScene::gridStartPos() const
{
    const qreal width = availableScreenRect().width();
    enum { MinimumMargin = 10 };
    const int columns = width / (itemSize().width() + MinimumMargin);
    const QSizeF item = itemSize();
    const qreal margin = (width - (columns * item.width())) / (columns + 1);
    QPointF p = QPointF(margin, margin);
    return p;
}

int GraphicsScene::layoutModeForKey(int key, Qt::KeyboardModifiers modifiers) const
{
    int mode = LayoutModeCount;
#if 0
    if (key >= Qt::Key_F1 && key <= Qt::Key_F12) {
        mode = static_cast<LayoutMode>(key - Qt::Key_F1);
    } else
#endif
    if (key >= Qt::Key_0 && key <= Qt::Key_9) {
        mode = key - Qt::Key_0;
        if(modifiers & Qt::ControlModifier)
            mode += 10;
    }
    return qMin<int>(LayoutModeCount, mode);
}

void GraphicsScene::computeSceneRect()
{
    QRectF rect(0, 0, 1, 1);
    foreach(const QGraphicsItem *item, items()) {
        rect |= item->sceneBoundingRect();
    }
    rect.setTopLeft(QPointF(0, 0));
    setSceneRect(rect);
}

void GraphicsScene::start()
{

    if (QCoreApplication::instance()->property("nowebpreload").toBool() == true) {
        onWebItemsLoaded();
        return;
    }

    QDir dir("/data/demo/Web");
    dir.setNameFilters(QStringList() << "*.url");
    const QList<QFileInfo> list = dir.entryInfoList();
    d.progressItem = new ProgressBarItem(list.size());
    QRectF r = availableScreenRect().adjusted(10, 0, -10, 0);
    r.setHeight(100);
    r.moveCenter(availableScreenRect().center());
    d.progressItem.data()->setGeometry(r);
    addItem(d.progressItem.data());


    foreach(const QFileInfo &fi, list) {
        WebItem *item = new WebItem(fi);
        item->resize(r.size());
        connect(item, SIGNAL(loadProgress(int)), d.progressItem.data(), SLOT(setValue(int)));
        d.cachedWebItems[fi.absoluteFilePath()] = item;
    }
    connect(d.progressItem.data(), SIGNAL(destroyed(QObject*)), this, SLOT(onWebItemsLoaded()));
    QTimer::singleShot(10000, d.progressItem.data(), SLOT(deleteLater()));

}
void GraphicsScene::onWebItemsLoaded()
{
    setDirectory(QDir("/data/demo"));
}

class UrlDialog : public QGraphicsProxyWidget
{
    Q_OBJECT
public:
    UrlDialog(QGraphicsWebView *webView)  : m_webView(webView)
    {
        Q_ASSERT(webView);
        GraphicsScene *scene = qobject_cast<GraphicsScene *>(webView->scene());
        Q_ASSERT(scene);

        blackItem = new QGraphicsRectItem(m_webView);
        blackItem->setRect(scene->sceneRect());
        blackItem->setPos(-m_webView->pos());
        blackItem->setBrush(QColor(0, 0, 0, 156));
        blackItem->setZValue(AlmostSpecialItemZ);

        QWidget *dialog = new QWidget;
        QHBoxLayout *layout = new QHBoxLayout(dialog);
        QLabel *label = new QLabel;
        label->setText(tr("Enter URL:"));
        layout->addWidget(label);
        lineEdit = new QLineEdit;
        layout->addWidget(lineEdit);
        lineEdit->setFocus();
        lineEdit->setText(m_webView->url().toString());
        lineEdit->setMinimumWidth(lineEdit->fontMetrics().width('X')*50);
        lineEdit->selectAll();
        lineEdit->installEventFilter(this);

        setWidget(dialog);
        setFlags(flags() | QGraphicsItem::ItemIsPanel | QGraphicsItem::ItemIsFocusable);
        setPanelModality(QGraphicsItem::SceneModal);
        setZValue(SpecialItemZ);

        setPos(QPoint((scene->visibleSceneRect().width() - dialog->sizeHint().width())/2, -dialog->sizeHint().height()));
        QPropertyAnimation *prop = new QPropertyAnimation(this, "pos");
        prop->setEndValue(QPoint((scene->visibleSceneRect().width() - dialog->sizeHint().width())/2, 0));
        prop->setDuration(scene->duration());
        prop->setEasingCurve(scene->curve());
        prop->start(QAbstractAnimation::DeleteWhenStopped);

        setFocus();
        lineEdit->setFocus();
        scene->addItem(this);
        connect(lineEdit, SIGNAL(returnPressed()), this, SLOT(handleEditingFinished()));
    }

    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = 0)
    {
        painter->fillRect(option->rect, Qt::gray);
        QGraphicsProxyWidget::paint(painter, option, widget);
    }

    bool eventFilter(QObject *object, QEvent *event)
    {
        if (object == lineEdit && event->type() == QEvent::KeyPress) {
            QKeyEvent *kev = static_cast<QKeyEvent *>(event);
            if (
#ifdef FRIENDLY_KEYS
                kev->key() == Qt::Key_Escape ||
#endif
                (kev->key() == Qt::Key_Left && kev->modifiers() == Qt::ControlModifier)) {
                close();
                return true;
            }
        }

        return QGraphicsProxyWidget::eventFilter(object, event);
    }
private slots:
    void handleEditingFinished()
    {
        m_webView->load(QUrl::fromUserInput(lineEdit->text()));
        close();
    }

    void close()
    {
        QGraphicsScene *scene = this->scene();
        scene->removeItem(this);
        scene->setFocusItem(m_webView);
        scene->removeItem(blackItem);
        delete blackItem;
        deleteLater();
    }

private:
    QGraphicsRectItem *blackItem;
    QLineEdit *lineEdit;
    QGraphicsWebView *m_webView;
};

#include "scene.moc"

void GraphicsScene::queryUrlDialog()
{
    Q_ASSERT(qobject_cast<QGraphicsWebView *>(currentItem()));

    new UrlDialog(qobject_cast<QGraphicsWebView *>(currentItem()));
}

GraphicsScene *GraphicsScene::inst = 0;
GraphicsScene * GraphicsScene::instance()
{
    return inst;
}

CoverFlowManager::CoverFlowManager(int side, GraphicsScene *scene)
    : QObject(scene)
{
    d.scene = scene;
    Q_ASSERT(d.scene);
    const QRectF r = d.scene->visibleSceneRect();
    d.itemSize = QSize(side, side);
    d.cachedPolygons.clear();
    const int s = d.scene->d.items.size();
    for (int i=-s; i<s; ++i) {
        initPolygons(i);
    }

#if 0
    // ### TODO Create proper semi-transparent reflections.
    const QColor background(Qt::black);
//    setBackgroundBrush(background);
    QLinearGradient gradient(0, 0, 0, side);
    gradient.setColorAt(1.0, QColor(background.red(), background.green(), background.blue(), 120));
    gradient.setColorAt(0.4, background);
    gradient.setColorAt(0.0, background);

    foreach(QGraphicsWidget *widget, d.scene->d.items) {
        ViewItem *item = qgraphicsitem_cast<ViewItem*>(widget);
        Q_ASSERT(item);
        QPixmap pixmap = item->pixmap();
        QPainter p(&pixmap);
        p.fillRect(pixmap.rect(), gradient);

// //        const QPixmap mirrored = QPixmap::fromImage(image.mirrored(false, true));
        ReflectionItem *reflection = new ReflectionItem(pixmap);
//        reflection->setProperty("index", i);
        d.reflections.append(reflection);
        d.scene->addItem(reflection);
//         ++i;
    }
#endif
    d.current = d.scene->d.items.indexOf(d.scene->currentItem());
}

CoverFlowManager::~CoverFlowManager()
{
    if (GraphicsScene::instance()) {
        const QPolygonF null;
        foreach(QGraphicsWidget *widget, d.scene->d.items) {
            ViewItem *item = qgraphicsitem_cast<ViewItem*>(widget);
            Q_ASSERT(item);
            item->setPolygon(null);
        }
    }
}

void CoverFlowManager::keyPressEvent(QKeyEvent *e)
{
    e->ignore();
    if (!e->modifiers()) {
        if (e->key() == Qt::Key_Right) {
            if (d.current > 0)
                animateTo(d.current - 1);
            e->accept();
        } else if (e->key() == Qt::Key_Left) {
            if (d.current + 1 < d.scene->d.items.size())
                animateTo(d.current + 1);
            e->accept();
        }
    }
}

void CoverFlowManager::setCurrentIndex(int idx)
{
    const int itemCount = d.scene->d.items.size();
    d.current = idx;
    for (int i=0; i<itemCount; ++i) {
        const QPair<QPolygonF, QPolygonF> pair = initPolygons(i);

        ViewItem *cover = qgraphicsitem_cast<ViewItem*>(d.scene->d.items.at(i));
        Q_ASSERT(cover);
        cover->setZValue(1000 - qAbs(i - d.current));
        cover->setPolygon(pair.first);

        //ReflectionItem *reflection = d.reflections.at(i);
        //reflection->setZValue(1000 - qAbs(i - d.current));
        //reflection->setPolygon(pair.second);
    }
}

QParallelAnimationGroup *CoverFlowManager::createAnimationTo(int idx)
{
    QParallelAnimationGroup *group = new QParallelAnimationGroup(this);
    const int duration = GraphicsScene::instance()->duration();
    const QEasingCurve curve = GraphicsScene::instance()->curve();
    enum { Duration = 500 };
    QPauseAnimation *pause = new QPauseAnimation(duration / 2);
//    connect(pause, SIGNAL(finished()), this, SLOT(updateZValues()));
    group->addAnimation(pause);
//    connect(group, SIGNAL(destroyed(QObject*)), this, SLOT(updateZValues()));

    d.current = idx;
    updateZValues();

    const int itemCount = d.scene->d.items.size();
    for (int i=0; i<itemCount; ++i) {
        const QPair<QPolygonF, QPolygonF> pair = initPolygons(i);

        ViewItem *cover = qgraphicsitem_cast<ViewItem*>(d.scene->d.items.at(i));
        QPropertyAnimation *animation = new QPropertyAnimation(cover, "polygon");
        animation->setDuration(duration);
        animation->setEasingCurve(curve);
        animation->setEndValue(qVariantFromValue<QPolygonF>(pair.first));
//        cover->setZValue(1000 - qAbs(i - d.current));
        group->addAnimation(animation);

        //ReflectionItem *reflection = d.reflections.at(i);
        //animation = new QPropertyAnimation(reflection, "polygon");
        //animation->setDuration(Duration);
        //animation->setEndValue(qVariantFromValue<QPolygonF>(pair.second));
        //animation->setDuration(duration);
        //animation->setEasingCurve(curve);
//        reflection->setZValue(1000 - qAbs(i - d.current));
        //group->addAnimation(animation);
    }
    return group;
}


void CoverFlowManager::animateTo(int idx)
{
    if (d.animationGroup)
        delete d.animationGroup.data();
    d.animationGroup = createAnimationTo(idx);
    d.animationGroup.data()->start(QAbstractAnimation::DeleteWhenStopped);
    d.scene->setCurrentItem(d.scene->d.items.at(idx));
}

QSize CoverFlowManager::itemSize() const
{
    return d.itemSize;
}

void CoverFlowManager::updateZValues()
{
    const int itemCount = d.scene->d.items.size();
    for (int i=0; i<itemCount; ++i) {
        const qreal z = 1000 - qAbs(i - d.current);
        d.scene->d.items.at(i)->setZValue(z);
        //d.reflections.at(i)->setZValue(z);
    }
}

QRectF CoverFlowManager::centerRect() const
{
    const QSize viewItemSize = itemSize();
    QRectF r(0, 0, viewItemSize.width(), viewItemSize.height());
    const QRectF sr = d.scene->visibleSceneRect();
    r.moveCenter(QPointF(sr.center().x(), qMax<qreal>(10.0, sr.top() + (sr.height() * .33))));
    return r;
}

QPair<QPolygonF, QPolygonF> CoverFlowManager::initPolygons(int idx) const
{
    const int offset = idx - d.current;
    QPair<QPolygonF, QPolygonF> &pair = d.cachedPolygons[offset];
    if (pair.first.isEmpty()) {
        const QRectF sr = d.scene->visibleSceneRect();
        const QRectF r = centerRect();
        pair.first.reserve(4);
        pair.second.reserve(4);
        if (offset == 0) {
            pair.first.append(r.topLeft());
            pair.first.append(r.topRight());
            pair.first.append(r.bottomRight());
            pair.first.append(r.bottomLeft());
        } else {
            static const qreal firstOffset = .2;
            static const qreal widthMultiplier = .667;
            static const qreal secondOffset = .05;
            const QSize viewItemSize = itemSize();
            if (offset > 0) {
                pair.first.append(QPointF(0, qreal(viewItemSize.height()) * firstOffset));
                pair.first.append(QPointF(qreal(viewItemSize.width()) * widthMultiplier, qreal(viewItemSize.height()) * secondOffset));
                pair.first.append(QPointF(qreal(viewItemSize.width()) * widthMultiplier, qreal(viewItemSize.height()) - qreal(viewItemSize.height()) * secondOffset));
                pair.first.append(QPointF(0, qreal(viewItemSize.height()) - qreal(viewItemSize.height()) * firstOffset));
                pair.first.translate(r.x(), r.y());
                pair.first.translate(qreal(viewItemSize.width()), 0);
                pair.first.translate(stackingWidth() * offset - 1, 0);
            } else {
                pair.first.append(QPointF(0, qreal(viewItemSize.height()) * secondOffset));
                pair.first.append(QPointF(qreal(viewItemSize.width()) * widthMultiplier, qreal(viewItemSize.height()) * firstOffset));
                pair.first.append(QPointF(qreal(viewItemSize.width()) * widthMultiplier, qreal(viewItemSize.height()) - qreal(viewItemSize.height()) * firstOffset));
                pair.first.append(QPointF(0, qreal(viewItemSize.height()) - qreal(viewItemSize.height()) * secondOffset));
                pair.first.translate(r.x(), r.y());
                pair.first.translate(-qreal(viewItemSize.width() * widthMultiplier), 0);
                pair.first.translate(-(stackingWidth() * qAbs(offset) - 1), 0);
            }
        }
        const QRectF boundingRect = pair.first.boundingRect();
        const QPointF center = boundingRect.center();
        QTransform mirror;
        mirror.translate(0, boundingRect.height());
        mirror.translate(center.x(), center.y());
        mirror.scale(1, -1);
        mirror.translate(-center.x(), -center.y());
        pair.second = mirror.map(pair.first);

        if (offset > 0) {
            pair.second[3] = pair.first[3];
        } else if (offset < 0) {
            pair.second[2] = pair.first[2];
        }
    }
    return pair;
}

qreal CoverFlowManager::stackingWidth() const
{
    return qreal(itemSize().width()) * .15;
}

