/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** All rights reserved.
** Contact: Nokia Corporation (qt-info@nokia.com)
**
** This file is part of the QtCore module of the Qt Toolkit.
**
** $QT_BEGIN_LICENSE:LGPL$
** No Commercial Usage
** This file contains pre-release code and may not be distributed.
** You may use this file in accordance with the terms and conditions
** contained in the Technology Preview License Agreement accompanying
** this package.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights.  These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** If you have questions regarding the use of this file, please contact
** Nokia at qt-info@nokia.com.
**
**
**
**
**
**
**
**
** $QT_END_LICENSE$
**
****************************************************************************/
#ifdef DEMO_GLVIEWPORT
#include <QGLWidget>
#endif
#include <QtGui>
#include <QtCore/qstate.h>

class Pixmap : public QObject, public QGraphicsPixmapItem
{
    Q_OBJECT
    Q_PROPERTY(QPointF pos READ pos WRITE setPos)
public:
    Pixmap(const QPixmap &pix)
        : QObject(), QGraphicsPixmapItem(pix)
    {
    }
};

class Trigger : public QGraphicsWidget
{
    Q_OBJECT
public:
    Trigger(QGraphicsItem *p)
        : QGraphicsWidget(p) { /**/ }
signals:
    void pressed();
};

class RandomTrigger : public QObject
{
    Q_OBJECT
public:
    RandomTrigger(QObject *p = 0)
        : QObject(p)
    {
        startTimer(2000);
    }
    void setTriggerList(QList<Trigger*> triggerList)
    {
        this->triggerList = triggerList;
    }
protected:
    void timerEvent(QTimerEvent * event)
    {
        Q_UNUSED(event)
            int index = qrand()%triggerList.count();
        QMetaObject::invokeMethod(triggerList[index], "pressed");
    }
private:
    QList<Trigger*> triggerList;
};

class Widget : public QGraphicsView
{
public:
    Widget()
        : QGraphicsView()
    {
        QPixmap kineticPix(":/images/kinetic.png");

        QGraphicsScene *scene = new QGraphicsScene(-350, -350, 700, 700);

        QList<Pixmap *> items;
        for (int i = 0; i < 64; ++i) {
            Pixmap *item = new Pixmap(kineticPix);
            item->setOffset(-kineticPix.width()/2, -kineticPix.height()/2);
            item->setZValue(i);
            items << item;
            scene->addItem(item);
        }

        // Triggers
        QGraphicsItem *parentItem = new QGraphicsRectItem(0, scene);
        Trigger *ellipseTrigger = new Trigger(parentItem);
        Trigger *figure8Trigger = new Trigger(parentItem);
        Trigger *randomTrigger = new Trigger(parentItem);
        Trigger *tiledTrigger = new Trigger(parentItem);
        Trigger *centeredTrigger = new Trigger(parentItem);

        QList<Trigger*> triggerList;
        triggerList << ellipseTrigger
                    << figure8Trigger
                    << randomTrigger
                    << tiledTrigger
                    << centeredTrigger;

        RandomTrigger *randomImpulse = new RandomTrigger(this);
        randomImpulse->setTriggerList(triggerList);

        ellipseTrigger->setPos(-100, -100);
        figure8Trigger->setPos(100, -100);
        randomTrigger->setPos(0, 0);
        tiledTrigger->setPos(-100, 100);
        centeredTrigger->setPos(100, 100);

        //scene->addItem(parentItem);
        parentItem->scale(0.75, 0.75);
        parentItem->setPos(200, 200);
        parentItem->setZValue(65);

        parentItem->hide();

        // States
        QState *rootState = new QState;
        QState *ellipseState = new QState(rootState);
        QState *figure8State = new QState(rootState);
        QState *randomState = new QState(rootState);
        QState *tiledState = new QState(rootState);
        QState *centeredState = new QState(rootState);

        // Values
        for (int i = 0; i < items.count(); ++i) {
            Pixmap *item = items.at(i);
            // Ellipse
            ellipseState->assignProperty(item, "pos",
                                         QPointF(cos((i / 63.0) * 6.28) * 250,
                                                 sin((i / 63.0) * 6.28) * 250));

            // Figure 8
            figure8State->assignProperty(item, "pos",
                                         QPointF(sin((i / 63.0) * 6.28) * 250,
                                                 sin(((i * 2)/63.0) * 6.28) * 250));

            // Random
            randomState->assignProperty(item, "pos",
                                        QPointF(-250 + qrand() % 500,
                                                -250 + qrand() % 500));

            // Tiled
            tiledState->assignProperty(item, "pos",
                                       QPointF(((i % 8) - 4) * kineticPix.width() + kineticPix.width() / 2,
                                              ((i / 8) - 4) * kineticPix.height() + kineticPix.height() / 2));

            // Centered
            centeredState->assignProperty(item, "pos", QPointF());
        }

        // Ui
        setScene(scene);

#ifdef DEMO_GLVIEWPORT
        setViewport(new QGLWidget());
#endif
#ifdef DEMO_GL
        setViewportUpdateMode(QGraphicsView::FullViewportUpdate);
        setRenderHints(QPainter::Antialiasing | QPainter::SmoothPixmapTransform);
#else
        setViewportUpdateMode(QGraphicsView::BoundingRectViewportUpdate);
#endif
        setBackgroundBrush(Qt::black);
        setWindowTitle(QT_TRANSLATE_NOOP(QGraphicsView, "Animated Tiles"));
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        setFrameStyle(0);

        QStateMachine *states = new QStateMachine(this);
        states->addState(rootState);
        states->setInitialState(rootState);
        rootState->setInitialState(centeredState);

        QParallelAnimationGroup *group = new QParallelAnimationGroup;
        for (int i = 0; i < items.count(); ++i) {
            QPropertyAnimation *anim = new QPropertyAnimation(items[i], "pos");
            anim->setDuration(750 + i * 25);
            anim->setEasingCurve(QEasingCurve::InOutBack);
            group->addAnimation(anim);
        }

        QAbstractTransition *trans = rootState->addTransition(ellipseTrigger, SIGNAL(pressed()), ellipseState);
        trans->addAnimation(group);

        trans = rootState->addTransition(figure8Trigger, SIGNAL(pressed()), figure8State);
        trans->addAnimation(group);

        trans = rootState->addTransition(randomTrigger, SIGNAL(pressed()), randomState);
        trans->addAnimation(group);

        trans = rootState->addTransition(tiledTrigger, SIGNAL(pressed()), tiledState);
        trans->addAnimation(group);

        trans = rootState->addTransition(centeredTrigger, SIGNAL(pressed()), centeredState);
        trans->addAnimation(group);

        states->start();
    }

protected:
    void resizeEvent(QResizeEvent *event)
    {
        QGraphicsView::resizeEvent(event);
        if (QCoreApplication::instance()->property("fitInView").toBool()) {
            fitInView(sceneRect(), Qt::KeepAspectRatio);
        }
    }
};

#include "../plugin.h"

DEFINE_PLUGIN_FUNCS(animatedtitles, "Animated Tiles", Widget);

#include "main.moc"
